#include <stdio.h>
#include <math.h>
#include "opcode.h"

typedef unsigned int UINT;

union REG 
{
	UINT u;
	int i;
	float f;
};

UINT xor(UINT a, UINT b){
  return ~a & b | a & ~b;
}


UINT pc;
UINT inst;
union REG reg[32];

#define INST_MEM_SIZE 0x100000
#define DATA_MEM_SIZE 0x100000

UINT *inst_mem;
UINT *data_mem;
char **inst_str;

void calc();
void show();

int SHOW_STATE = 0;		//-s V~[^̂PXebvƂ̏Ԃ 0:\Ȃ 1:\
int IOFORMAT_X = 0;		//-x o͂̌` 0:10i 1:16i

int main(int argc, char *argv[])
{
	FILE *fp;
	int i, j;
	
	fp = NULL;
	for(i = 1; i < argc; i++)
		if(argv[i][0] == '-')
			for(j = 1; argv[i][j] != '\0'; j++)
				switch(argv[i][j])
				{
				case 's':
					SHOW_STATE = 1;
					break;
				case 'x':
					IOFORMAT_X = 1;
					break;
				default:
					printf("Unknown option: -%c\n", argv[i][j]);
				}
		else
			fp = fopen(argv[i], "r");
	
	if(!fp)
	{
		printf("usage: %s [-sx] input_file\n",argv[0]);
		return 1;
	}
	
	inst_mem = (UINT *)malloc(sizeof(int) * INST_MEM_SIZE);
	data_mem = (UINT *)malloc(sizeof(int) * DATA_MEM_SIZE);
	inst_str = (char **)malloc(sizeof(char *) * INST_MEM_SIZE);
	
	for(i=0;!feof(fp);i++){
		inst_str[i] = (char *)malloc(sizeof(char) * 64);
		fscanf(fp, "0x%08X\t%[^\n]\n", inst_mem + i, inst_str[i]);
	}
	
	pc = 0;
	reg[0].u =  0;			// [WX^	
	reg[1].u  = 0;			// ֐ĂяõX^bN|C^
	reg[2].u  = 0x80014;		// ^v쐬p̃X^bN|C^
	reg[30].u = 0xC019F;		// z쐬p̃X^bN|C^
	
	calc();
	
	return 0;
}

void show()
{
	int i;
	char buf[8];
	
	printf("PC : %d\n", pc);
	printf("INST : 0x%08X\t%s\n", inst, inst_str[pc]);
	
	for(i=0;i<16;i++)
		printf("r%2d : 0x%08X\tr%2d : 0x%08X\n", i, reg[i].u, i+16, reg[i+16].u);
	
	printf("press 'y' to show next step, 'n' to exit step-mode\n");
	fgets(buf, 8, stdin);
	
	if(*buf == 'n')
	  SHOW_STATE = 0;
	else
	  SHOW_STATE = 1;
}

void calc()
{
	UINT op, rs, rt, rd, shft, fnct;
	int imm, target;

	inst = inst_mem[pc];
	if(SHOW_STATE)show();
	
	imm    = ((inst & 0x8000)    ? 0xFFFF << 16 : 0) | inst & 0xFFFF;
	op   = inst >> 26;
	rs   = inst >> 21 & 0x1F;
	rt   = inst >> 16 & 0x1F;
	rd   = inst >> 11 & 0x1F;
	shft = inst >> 6  & 0x1F;
	fnct = inst       & 0x3F;
	target = ((inst & 0x2000000) ? 0xFC   << 24 : 0) | inst & 0x3FFFFFF;
	
	switch(op)
	{
	/**********************************/
	/* ALU OPERATIONS                 */
	/**********************************/
	case OP_ALU:
		switch(fnct)
		{
		case FNCT_ADD:
			reg[rd].i = reg[rs].i + reg[rt].i;
			break;
		case FNCT_SUB:
			reg[rd].i = reg[rs].i - reg[rt].i;
			break;
		case FNCT_MUL:
			reg[rd].i = reg[rs].i * reg[rt].i;
			break;
		case FNCT_DIV:
			reg[rd].i = reg[rs].i / reg[rt].i;
			break;
		case FNCT_SLL:
			reg[rd].u = reg[rs].u << shft;
			break;
		case FNCT_SRL:
			reg[rd].u = reg[rs].u >> shft;
			break;
		case FNCT_SRR:
			reg[rd].i = reg[rs].i >> shft;
			break;
		case FNCT_AND:
			reg[rd].u = reg[rs].u & reg[rt].u;
			break;
		case FNCT_OR:
			reg[rd].u = reg[rs].u | reg[rt].u;
			break;
		case FNCT_XOR:
			reg[rd].u = xor(reg[rs].u, reg[rt].u);
			break;
		case FNCT_NOT:
			reg[rd].u = ~reg[rs].u;
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* IMMEDIATE OPERATIONS           */
	/**********************************/
	case OP_ADDI:
		reg[rt].i = reg[rs].i + imm;
		pc++;
		break;
	case OP_SUBI:
		reg[rt].i = reg[rs].i - imm;
		pc++;
		break;
	case OP_MULI:
		reg[rt].i = reg[rs].i * imm;
		pc++;
		break;
	case OP_DIVI:
		reg[rt].i = reg[rs].i / imm;
		pc++;
		break;
	case OP_ANDI:
		reg[rt].u = reg[rs].u & imm;
		pc++;
		break;
	case OP_ORI:
		reg[rt].u = reg[rs].u | imm;
		pc++;
		break;
	case OP_XORI:
		reg[rt].u = xor(reg[rs].u, (UINT)imm);
		pc++;
		break;
	
	/**********************************/
	/* FPU OPERATIONS                 */
	/**********************************/
	case OP_FPU:
		switch(fnct)
		{
		case FNCT_FADD:
			reg[rd].f = reg[rs].f + reg[rt].f;
			break;
		case FNCT_FSUB:
			reg[rd].f = reg[rs].f - reg[rt].f;
			break;
		case FNCT_FMUL:
			reg[rd].f = reg[rs].f * reg[rt].f;
			break;
		case FNCT_FDIV:
			reg[rd].f = reg[rs].f / reg[rt].f;
			break;
		case FNCT_ITOF:
			reg[rd].f = (float)reg[rs].i;
			break;
		case FNCT_FTOI:
			reg[rd].i = (int)reg[rs].f;
			break;
		case FNCT_FABS:
			reg[rd].f = fabsf(reg[rs].f);
			break;
		case FNCT_FNEG:
			reg[rd].f = -reg[rs].f;
			break;
		case FNCT_FLOOR:
			reg[rd].f = floor(reg[rs].f);
			break;
		case FNCT_FSQRT:
			reg[rd].f = (float)sqrt(reg[rs].f);	//sqrtf
			break;
		case FNCT_SIN:
			reg[rd].f = (float)sin(reg[rs].f);
			break;
		case FNCT_COS:
			reg[rd].f = (float)cos(reg[rs].f);
			break;
		case FNCT_ATAN:
			reg[rd].f = (float)atan(reg[rs].f);
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* SYSTEM OPERATIONS              */
	/**********************************/
	case OP_SYS:
		switch(fnct){
		case FNCT_NOP:
			break;
		case FNCT_HLT:
			printf("system halt...\n");
			return;
		case FNCT_RD:
			if(IOFORMAT_X)
				scanf("%X", &reg[rd].u);
			else
				scanf("%d", &reg[rd].u);
			break;
		case FNCT_WR:
			if(IOFORMAT_X)
				printf("0x%X\n", reg[rs].u);
			else
				printf("%d\n", reg[rs].u);
			break;
		case FNCT_FRD:
			scanf("%f", &reg[rd].f);
			break;
		case FNCT_BR:
			printf("break: %d\n", rs);
			show();
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* MEMORY OPERATIONS              */
	/**********************************/
	case OP_LOAD:{
		UINT addr = reg[rs].u + imm;
		if(addr < 0 || addr >= 0x100000)
		{
			printf("load\n");
			show();
		}}
		reg[rt].u = data_mem[reg[rs].u + imm];
		pc++;
		break;
	case OP_STORE:{
		UINT addr = reg[rs].u + imm;
		if(addr < 0 || addr >= 0x100000)
		{
			printf("store\n");
			show();
		}}
		data_mem[reg[rs].u + imm] = reg[rt].u;
		pc++;
		break;
	
	/**********************************/
	/* JUMP OPERATIONS                */
	/**********************************/
	case OP_JUMP:
		pc = pc + target;
		break;
	case OP_JAL:
		reg[31].u = pc + 1;		//1 or 4?
		pc = pc + (unsigned int)target;
		break;
	case OP_JR:
		pc = reg[rs].u;
		break;
	
	/**********************************/
	/* INT BRANCH OPERATIONS          */
	/**********************************/
	case OP_BL  :
		if(reg[rs].i <  reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BE  :                  
		if(reg[rs].i == reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BLE :                  
		if(reg[rs].i <= reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BGE :                  
		if(reg[rs].i >= reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BNE :                  
		if(reg[rs].i != reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BG  :                  
		if(reg[rs].i >  reg[rt].i) pc += imm; else pc ++;
		break;
	case OP_BLZ :
		if(reg[rs].i <  0) pc += imm; else pc ++;
		break;    
	case OP_BEZ :     
		if(reg[rs].i == 0) pc += imm; else pc ++;
		break;    
	case OP_BLEZ:     
		if(reg[rs].i <= 0) pc += imm; else pc ++;
		break;    
	case OP_BGEZ:     
		if(reg[rs].i >= 0) pc += imm; else pc ++;
		break;    
	case OP_BNEZ:     
		if(reg[rs].i != 0) pc += imm; else pc ++;
		break;    
	case OP_BGZ :     
		if(reg[rs].i >  0) pc += imm; else pc ++;
		break;
	/**********************************/
	/* FLOAT BRANCH OPERATIONS        */
	/**********************************/
	case OP_FBL  :
		if(reg[rs].f <  reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBE  :                 
		if(reg[rs].f == reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBLE :                 
		if(reg[rs].f <= reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBGE :                 
		if(reg[rs].f >= reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBNE :                 
		if(reg[rs].f != reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBG  :                 
		if(reg[rs].f >  reg[rt].f) pc += imm; else pc ++;
		break;
	case OP_FBLZ :
		if(reg[rs].f <  0) pc += imm; else pc ++;
		break;    
	case OP_FBEZ :     
		if(reg[rs].f == 0) pc += imm; else pc ++;
		break;    
	case OP_FBLEZ:     
		if(reg[rs].f <= 0) pc += imm; else pc ++;
		break;    
	case OP_FBGEZ:     
		if(reg[rs].f >= 0) pc += imm; else pc ++;
		break;    
	case OP_FBNEZ:     
		if(reg[rs].f != 0) pc += imm; else pc ++;
		break;    
	case OP_FBGZ :    
		if(reg[rs].f >  0) pc += imm; else pc ++;
		break;
	}
	
	calc();
}


